from dataclasses import dataclass
from typing import Any, List

from oaklib.resource import OntologyResource


@dataclass
class OntologyInterface:
    """
    An abstract parent for an ontology interface
    """
    #engine: Any = None  ## implementation object
    resource: OntologyResource = None
    strict: bool = False

    @classmethod
    def create(cls, resource: OntologyResource) -> "OntologyInterface":
        """
        Creates a new ontology interface from a resource

        :param resource:
        :return:
        """
        raise NotImplementedError

    def interfaces_implemented(self) -> List:
        """
        For any given instance of this interface, find all interfaces implemented

        :return:
        """
        return list(type(self).__bases__)



import itertools
from typing import Iterable, List

DEFAULT_CHUNK = 100


def chunk(iterable: Iterable, size=DEFAULT_CHUNK) -> Iterable[List]:
    """
    Get first N results of iterable

    https://stackoverflow.com/questions/8991506/iterate-an-iterator-by-chunks-of-n-in-python
    """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, size)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


def chunk_to_lists(iterable: Iterable, size=DEFAULT_CHUNK) -> Iterable[List]:
    """
    Get first N results of iterable

    https://stackoverflow.com/questions/8991506/iterate-an-iterator-by-chunks-of-n-in-python
    """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, size)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield list(itertools.chain((first_el,), chunk_it))



import logging
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Iterable, Type, Iterator, Union

import pronto
import sssom
from deprecated import deprecated
from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface, RELATIONSHIP_MAP, PRED_CURIE, ALIAS_MAP, \
    METADATA_MAP, PREFIX_MAP
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.interfaces.search_interface import SearchInterface, SearchConfiguration
from oaklib.interfaces.validator_interface import ValidatorInterface
from oaklib.interfaces.rdf_interface import RdfInterface
from oaklib.interfaces.relation_graph_interface import RelationGraphInterface
from oaklib.resource import OntologyResource
from oaklib.types import CURIE, SUBSET_CURIE
from oaklib.datamodels import obograph
from oaklib.datamodels.obograph import Edge, Graph
from oaklib.datamodels.vocabulary import LABEL_PREDICATE, IS_A, HAS_DBXREF, SCOPE_TO_SYNONYM_PRED_MAP, SKOS_CLOSE_MATCH
from pronto import Ontology, LiteralPropertyValue, ResourcePropertyValue, Term
from sssom.sssom_datamodel import MatchTypeEnum


@dataclass
class ProntoImplementation(ValidatorInterface, RdfInterface, RelationGraphInterface, OboGraphInterface, SearchInterface, MappingProviderInterface):
    """
    Pronto wraps local-file based ontologies in the following formats:

    - obo
    - obojson
    - owl rdf/xml

    To load a local file:

    .. code:: python

        >>> resource = OntologyResource(slug='go.obo', directory='input', local=True)
        >>> oi = ProntoImplementation.create(resource)

    To load from the OBO library:

    .. code:: python

        >>> resource = OntologyResource(local=False, slug='go.obo'))
        >>> oi = ProntoImplementation.create(resource)

    Currently this implementation implements most of the BaseOntologyInterface

    .. code:: python

        rels = oi.get_outgoing_relationships_by_curie('GO:0005773')
        for rel, parents in rels.items():
            print(f'  {rel} ! {oi.get_label_by_curie(rel)}')
                for parent in parents:
                    print(f'    {parent} ! {oi.get_label_by_curie(parent)}')


    """
    wrapped_ontology: Ontology = None

    def __post_init__(self):
        if self.wrapped_ontology is None:
            resource = self.resource
            if resource is None:
                ontology = Ontology()
            elif resource.local:
                ontology = Ontology(str(resource.local_path))
            else:
                ontology = Ontology.from_obo_library(resource.slug)
            self.wrapped_ontology = ontology

    @classmethod
    @deprecated('old style')
    def create(cls, resource: OntologyResource = None) -> "ProntoImplementation":
        return ProntoImplementation(resource=resource)

    def store(self, resource: OntologyResource = None) -> None:
        if resource is None:
            resource = self.resource
        ontology = self.wrapped_ontology
        if resource.local:
            with open(str(resource.local_path), 'wb') as f:
                ontology.dump(f, format=resource.format)
        else:
            raise NotImplementedError(f'Cannot dump to {resource}')

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: BasicOntologyInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def get_prefix_map(self) -> PREFIX_MAP:
        return {}

    def _entity(self, curie: CURIE):
        for r in self.wrapped_ontology.relationships():
            # see https://owlcollab.github.io/oboformat/doc/obo-syntax.html#4.4.1
            # pronto gives relations shorthand IDs for RO and BFO, as it is providing
            # oboformat as a level of abstraction. We want to map these back to the CURIEs
            if r.id == curie:
                return r
            if curie.startswith('RO:') or curie.startswith('BFO:'):
                if any(x for x in r.xrefs if x.id == curie):
                    return r
        if curie in self.wrapped_ontology:
            return self.wrapped_ontology[curie]
        else:
            return None

    def _create(self, curie: CURIE, exist_ok = True):
        if curie in self.wrapped_ontology:
            return self.wrapped_ontology[curie]
        else:
            return self.wrapped_ontology.create_term(curie)

    def _create_pred(self, curie: CURIE, exist_ok = True):
        if curie in self.wrapped_ontology:
            return self.wrapped_ontology[curie]
        else:
            return self.wrapped_ontology.create_relationship(curie)

    def all_entity_curies(self) -> Iterable[CURIE]:
        for t in self.wrapped_ontology.terms():
            yield t.id
        # note what Pronto calls "relationship" is actually "relationship type"
        for t in self.wrapped_ontology.relationships():
            yield t.id
        for t in self.wrapped_ontology.synonym_types():
            yield t.id

    def all_subset_curies(self) -> Iterable[CURIE]:
        subsets = set()
        for t in self.wrapped_ontology.terms():
            subsets.update(t.subsets)
        for subset in subsets:
            yield subset

    def curies_by_subset(self, subset: SUBSET_CURIE) -> Iterable[CURIE]:
        for t in self.wrapped_ontology.terms():
            if subset in t.subsets:
                yield t.id

    def get_label_by_curie(self, curie: CURIE) -> str:
        t = self._entity(curie)
        if t:
            return t.name
        else:
            if curie == IS_A:
                return 'subClassOf'
            else:
                return None

    def set_label_for_curie(self, curie: CURIE, label: str) -> bool:
        t = self._entity(curie)
        if t:
            curr = t.name
            if curr != label:
                t.name = label
                return True
            else:
                return False

    def get_curies_by_label(self, label: str) -> List[CURIE]:
        return [t.id for t in self.wrapped_ontology.terms() if t.name == label]

    def _get_pronto_relationship_type_curie(self, rel_type: pronto.Relationship) -> CURIE:
        for x in rel_type.xrefs:
            if x.id.startswith('BFO:') or x.id.startswith('RO:'):
                return x.id
        return rel_type.id


    def get_outgoing_relationships_by_curie(self, curie: CURIE, isa_only: bool = False) -> RELATIONSHIP_MAP:
        # See: https://github.com/althonos/pronto/issues/119
        term = self._entity(curie)
        if isinstance(term, Term):
            # only "Terms" in pronto have relationships
            rels = {IS_A: [p.id for p in term.superclasses(distance=1) if p.id != curie]}
            for rel_type, parents in term.relationships.items():
                pred = self._get_pronto_relationship_type_curie(rel_type)
                rels[pred] = [p.id for p in parents]
        else:
            rels = {}
        return rels

    def get_incoming_relationships_by_curie(self, curie: CURIE, isa_only: bool = False) -> RELATIONSHIP_MAP:
        term = self._entity(curie)
        if isinstance(term, Term):
            # only "Terms" in pronto have relationships
            rels = {IS_A: [p.id for p in term.subclasses(distance=1) if p.id != curie]}
            for xt in self.wrapped_ontology.terms():
                for rel_type, parents in xt.relationships.items():
                    pred = self._get_pronto_relationship_type_curie(rel_type)
                    for p in parents:
                        if curie == p.id:
                            if pred not in rels:
                                rels[pred] = []
                            rels[pred].append(xt.id)
        else:
            rels = {}
        return rels




    def create_entity(self, curie: CURIE, label: str = None, relationships: RELATIONSHIP_MAP = None) -> CURIE:
        ont = self.wrapped_ontology
        t = ont.create_term(curie)
        t.name = label
        for pred, fillers in relationships.items():
            for filler in fillers:
                self.add_relationship(curie, pred, filler)
        return curie

    def add_relationship(self, curie: CURIE, predicate: PRED_CURIE, filler: CURIE):
        t = self._entity(curie)
        filler_term = self._create(filler)
        if predicate == IS_A:
            t.superclasses().add(filler_term)
        else:
            predicate_term = self._create_pred(predicate)
            if predicate_term not in t.relationships.keys():
                t.relationships[predicate_term] = []
            t.relationships[predicate_term].add(filler_term)

    def get_definition_by_curie(self, curie: CURIE) -> str:
        return self._entity(curie).definition

    def alias_map_by_curie(self, curie: CURIE) -> ALIAS_MAP:
        t = self._entity(curie)
        if t is None:
            return {}
        m = defaultdict(list)
        m[LABEL_PREDICATE] = [t.name]
        for s in t.synonyms:
            scope = s.scope.upper()
            if scope in SCOPE_TO_SYNONYM_PRED_MAP:
                pred = SCOPE_TO_SYNONYM_PRED_MAP[scope]
            else:
                raise ValueError(f'Unknown scope: {scope}')
            m[pred].append(s.description)
        return m

    def get_simple_mappings_by_curie(self, curie: CURIE) -> RELATIONSHIP_MAP:
        m = defaultdict(list)
        t = self._entity(curie)
        if t is None:
            return m
        for s in t.xrefs:
            m[HAS_DBXREF].append(s.id)
        for s in t.annotations:
            # TODO: less hacky
            if s.property.startswith('skos'):
                if isinstance(s, LiteralPropertyValue):
                    v = s.literal
                    m[s.property].append(v)
                elif isinstance(s, ResourcePropertyValue):
                    v = self.uri_to_curie(s.resource)
                else:
                    v = None
                if v:
                    m[s.property].append(v)
        return m

    def metadata_map_by_curie(self, curie: CURIE) -> METADATA_MAP:
        t = self._entity(curie)
        m = defaultdict(list)
        for ann in t.annotations:
            p = ann.property
            if isinstance(ann, LiteralPropertyValue):
                m[ann.property].append(ann.literal)
            elif isinstance(ann, ResourcePropertyValue):
                m[ann.property].append(ann.resource)
        return m

    def create_subontology(self, curies: List[CURIE]) -> "ProntoImplementation":
        subontology = Ontology()
        for curie in curies:
            t = self._entity(curie)
            subontology.create_term(curie)
            t2 = subontology[curie]
            t2.name = t.name
            # TODO - complete object
        return ProntoImplementation(wrapped_ontology=subontology)

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: MappingsInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def get_sssom_mappings_by_curie(self, curie: Union[str, CURIE]) -> Iterator[sssom.Mapping]:
        t = self._entity(curie)
        if not t:
            return
            #raise ValueError(f'No such entity: {curie}')
        for x in t.xrefs:
            yield sssom.Mapping(subject_id=curie,
                                predicate_id=SKOS_CLOSE_MATCH,
                                object_id=x.id,
                                match_type=MatchTypeEnum.Lexical)


    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: OboGraphInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def node(self, curie: CURIE, strict=False) -> obograph.Node:
        t = self._entity(curie)
        if t is None:
            return obograph.Node(id=curie)
        else:
            meta = obograph.Meta()
            if t.definition:
                meta.definition = obograph.DefinitionPropertyValue(val=t.definition)
            if isinstance(t, pronto.Relationship):
                t_id = self._get_pronto_relationship_type_curie(t)
            else:
                t_id = t.id
            if isinstance(t, pronto.Relationship):
                for x in t.xrefs:
                    if x.id.startswith('RO:') or x.id.startswith('BFO:'):
                        t_id = x.id
            #for s in t.synonyms:
            #    meta.synonyms.append(obograph.SynonymPropertyValue(val=s.description,
            #                                                       scope=s.scope.lower(),
            #                                                      xrefs=[x.id for x in s.xrefs]))
            return obograph.Node(id=t_id,
                                 label=t.name,
                                 meta=meta)

    def as_obograph(self) -> Graph:
        nodes = [self.node(curie) for curie in self.all_entity_curies()]
        edges = [Edge(sub=r[0], pred=r[1], obj=r[2]) for r in self.all_relationships()]
        return Graph(id='TODO', nodes=nodes, edges=edges)

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: SearchInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def basic_search(self, search_term: str, config: SearchConfiguration = None) -> Iterable[CURIE]:
        if config == None:
            config = SearchConfiguration()
        matches = []
        for t in self.wrapped_ontology.terms():
            if t.name and search_term in t.name:
                matches.append(t.id)
                logging.info(f'Name match to {t.id}')
                continue
            if config.include_aliases:
                for syn in t.synonyms:
                    if search_term in syn.description:
                        logging.info(f'Syn match to {t.id}')
                        matches.append(t.id)
                        continue
        return matches




from abc import ABC
from enum import Enum
from typing import Dict, List, Tuple, Iterable, Optional

# TODO: add funowl to dependencies
from funowl import Ontology, Axiom, Class, SubClassOf, EquivalentClasses
from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.types import CURIE, LABEL, URI
from rdflib import Graph


class OwlProfile(Enum):
    EL = "EL"
    DL = "DL"
    OWL_FULL = "OWL-Full"
    RL = "RL"
    QC = "QL"

@dataclass
class ReasonerConfiguration:
    reasoner: str = None
    reasoner_version: str = None
    implements_profiles: List[OwlProfile] = None

class OwlInterface(BasicOntologyInterface, ABC):
    """
    presents ontology as an OWL ontology using an OWL datamodel

    We leverage the funowl datamodel

    In future there will be a variety of implementations:

    - funowl
    - owlery
    - owllink servers
    - robot/owlapi via py4j
    """

    def owl_ontology(self) -> Ontology:
        raise NotImplementedError

    def axioms(self, reasoner: Optional[ReasonerConfiguration] = None) -> Iterable[Axiom]:
        raise NotImplementedError

    def subclass_axioms(self, subclass: CURIE = None, superclass: CURIE = None, reasoner: Optional[ReasonerConfiguration] = None) -> Iterable[SubClassOf]:
        """
        Gets all SubClassOf axioms matching criterion

        :param subclass: if specified, constrains to axioms where this is the subclass
        :param superclass: if specified, constrains to axioms where this is the superclass
        :param reasoner:
        :return:
        """
        raise NotImplementedError

    def equivalence_axioms(self, about: CURIE = None, references: CURIE = None, reasoner: Optional[ReasonerConfiguration] = None):
        raise NotImplementedError

    def owl_classes(self) -> Iterable[Class]:
        raise NotImplementedError

    def owl_individuals(self) -> Iterable[Class]:
        raise NotImplementedError

    def is_satisfiable(self, curie: CURIE) -> bool:
        raise NotImplementedError

    def reasoner_configurations(self) -> List[ReasonerConfiguration]:
        raise NotImplementedError




import csv
import sys
from dataclasses import dataclass, field
from typing import Any, Dict, Union

from linkml_runtime import CurieNamespace
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.yamlutils import YAMLRoot
from oaklib.io.streaming_writer import StreamingWriter


def _keyval(x: Any) -> str:
    if isinstance(x, CurieNamespace):
        return str(x.curie())
    #if isinstance(x, EnumDefinitionImpl):
    #    if x.curie:
    #        return str(x.curie)
    return str(x)


@dataclass
class StreamingCsvWriter(StreamingWriter):
    """
    A writer that streams CSV/TSV output
    """

    header_emitted: bool = None
    delimiter: str = '\t'
    writer: csv.DictWriter = None

    def emit(self, obj: Union[YAMLRoot, Dict]):
        if isinstance(obj, dict):
            obj_as_dict = obj
        else:
            obj_as_dict = vars(obj)
        if self.writer is None:
            self.writer = csv.DictWriter(self.file, delimiter=self.delimiter, fieldnames=list(obj_as_dict))
            self.writer.writeheader()
        self.writer.writerow({k: _keyval(v) for k, v in obj_as_dict.items()})





# Auto generated from search_results.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-04-05T13:39:03
# Schema: search-results
#
# id: https://w3id.org/linkml/search_results
# description: A datamodel for representing the results of ontology search
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, Integer, String
from linkml_runtime.utils.metamodelcore import Bool

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
PROV = CurieNamespace('prov', 'http://www.w3.org/ns/prov#')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SH = CurieNamespace('sh', 'https://w3id.org/shacl/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
SR = CurieNamespace('sr', 'https://w3id.org/linkml/search_results/')
SSSOM = CurieNamespace('sssom', 'http://w3id.org/sssom/')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = SR


# Types
class Position(Integer):
    type_class_uri = XSD.integer
    type_class_curie = "xsd:integer"
    type_name = "Position"
    type_model_uri = SR.Position


# Class references



@dataclass
class SearchResultSet(YAMLRoot):
    """
    A collection of annotation results
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SR.SearchResultSet
    class_class_curie: ClassVar[str] = "sr:SearchResultSet"
    class_name: ClassVar[str] = "SearchResultSet"
    class_model_uri: ClassVar[URIRef] = SR.SearchResultSet

    results: Optional[Union[Union[dict, "SearchResult"], List[Union[dict, "SearchResult"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.results, list):
            self.results = [self.results] if self.results is not None else []
        self.results = [v if isinstance(v, SearchResult) else SearchResult(**as_dict(v)) for v in self.results]

        super().__post_init__(**kwargs)


@dataclass
class SearchResult(YAMLRoot):
    """
    An individual text annotation
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SR.SearchResult
    class_class_curie: ClassVar[str] = "sr:SearchResult"
    class_name: ClassVar[str] = "SearchResult"
    class_model_uri: ClassVar[URIRef] = SR.SearchResult

    object_id: Optional[str] = None
    object_label: Optional[str] = None
    object_source: Optional[str] = None
    matches_full_search_term: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.object_id is not None and not isinstance(self.object_id, str):
            self.object_id = str(self.object_id)

        if self.object_label is not None and not isinstance(self.object_label, str):
            self.object_label = str(self.object_label)

        if self.object_source is not None and not isinstance(self.object_source, str):
            self.object_source = str(self.object_source)

        if self.matches_full_search_term is not None and not isinstance(self.matches_full_search_term, Bool):
            self.matches_full_search_term = Bool(self.matches_full_search_term)

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass

slots.searchResultSet__results = Slot(uri=SR.results, name="searchResultSet__results", curie=SR.curie('results'),
                   model_uri=SR.searchResultSet__results, domain=None, range=Optional[Union[Union[dict, SearchResult], List[Union[dict, SearchResult]]]])

slots.searchResult__object_id = Slot(uri=SSSOM.object_id, name="searchResult__object_id", curie=SSSOM.curie('object_id'),
                   model_uri=SR.searchResult__object_id, domain=None, range=Optional[str])

slots.searchResult__object_label = Slot(uri=SSSOM.object_label, name="searchResult__object_label", curie=SSSOM.curie('object_label'),
                   model_uri=SR.searchResult__object_label, domain=None, range=Optional[str])

slots.searchResult__object_source = Slot(uri=SSSOM.object_source, name="searchResult__object_source", curie=SSSOM.curie('object_source'),
                   model_uri=SR.searchResult__object_source, domain=None, range=Optional[str])

slots.searchResult__matches_full_search_term = Slot(uri=SR.matches_full_search_term, name="searchResult__matches_full_search_term", curie=SR.curie('matches_full_search_term'),
                   model_uri=SR.searchResult__matches_full_search_term, domain=None, range=Optional[Union[bool, Bool]])



from abc import ABC
from typing import Dict, List, Iterable

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.resource import OntologyResource
from oaklib.types import CURIE


class DifferInterface(BasicOntologyInterface, ABC):
    """
    Generates descriptions of differences

    TBD: low level diffs vs high level

     See `KGCL <https://github.com/cmungall/knowledge-graph-change-language>`_
    """

    def diff(self, left_ontology_id: CURIE, right_ontology_id: CURIE) -> str:
        """
        Diffs two ontologies - both must be ontologies wrapped by the current implementation

        :param left_ontology_id:
        :param right_ontology_id:
        :return: TBD KGCL?
        """



from dataclasses import dataclass

from oaklib.implementations.sparql.sparql_implementation import SparqlImplementation
from oaklib.interfaces.owl_interface import OwlInterface


@dataclass
class OwleryImplementation(OwlInterface, SparqlImplementation):
    """
    Wraps an owlery endpoint

     See `<https://github.com/phenoscape/owlery>`_

    .. warning ::

      Not implemented
    """











from sqlalchemy import Column, Index, Table, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.sql.sqltypes import *
from sqlalchemy.orm import declarative_base
from sqlalchemy.ext.associationproxy import association_proxy

Base = declarative_base()
metadata = Base.metadata


class SubjectPrefix(Base):
    """
    This may move to another module as it is generally useful
    """
    __tablename__ = 'subject_prefix'
    
    subject = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subject_prefix(subject={self.subject},value={self.value},)"
        
    
        
    


class TextualTransformation(Base):
    """
    Represents a transformation of a subject text value, e.g. lemmatization
    """
    __tablename__ = 'textual_transformation'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"textual_transformation(subject={self.subject},predicate={self.predicate},value={self.value},)"
        
    
        
    


class ProcessedStatement(Base):
    """
    A statement that is enhanced with a processed/transformed text value
    """
    __tablename__ = 'processed_statement'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    transformation_predicate = Column(Text(), primary_key=True)
    transformed_value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"processed_statement(subject={self.subject},predicate={self.predicate},value={self.value},transformation_predicate={self.transformation_predicate},transformed_value={self.transformed_value},)"
        
    
        
    


class Match(Base):
    """
    TODO: Reuse SSSOM here
    """
    __tablename__ = 'match'
    
    subject_id = Column(Text(), primary_key=True)
    subject_label = Column(Text(), primary_key=True)
    subject_source = Column(Text(), primary_key=True)
    subject_match_field = Column(Text(), primary_key=True)
    subject_preprocessing = Column(Text(), primary_key=True)
    object_id = Column(Text(), primary_key=True)
    object_label = Column(Text(), primary_key=True)
    object_source = Column(Text(), primary_key=True)
    object_match_field = Column(Text(), primary_key=True)
    object_preprocessing = Column(Text(), primary_key=True)
    match_string = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"match(subject_id={self.subject_id},subject_label={self.subject_label},subject_source={self.subject_source},subject_match_field={self.subject_match_field},subject_preprocessing={self.subject_preprocessing},object_id={self.object_id},object_label={self.object_label},object_source={self.object_source},object_match_field={self.object_match_field},object_preprocessing={self.object_preprocessing},match_string={self.match_string},)"
        
    
        
    


class Prefix(Base):
    """
    Maps CURIEs to URIs
    """
    __tablename__ = 'prefix'
    
    prefix = Column(Text(), primary_key=True)
    base = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"prefix(prefix={self.prefix},base={self.base},)"
        
    
        
    


class Statements(Base):
    """
    Represents an RDF triple
    """
    __tablename__ = 'statements'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"statements(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    


class Node(Base):
    """
    The basic unit of representation in an RDF or OWL graph
    """
    __tablename__ = 'node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"node(id={self.id},)"
        
    
        
    


class RdfLevelSummaryStatistic(Base):
    """
    Abstract grouping for views/classes that provide some kind of count summary about an individual element
    """
    __tablename__ = 'rdf_level_summary_statistic'
    
    element = Column(Text(), primary_key=True)
    count_value = Column(Integer(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_level_summary_statistic(element={self.element},count_value={self.count_value},)"
        
    
        
    


class RelationGraphConstruct(Base):
    """
    A construct used as part of a Relation Graph
    """
    __tablename__ = 'relation_graph_construct'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"relation_graph_construct(subject={self.subject},predicate={self.predicate},object={self.object},)"
        
    
        
    


class OwlComplexAxiom(Base):
    """
    An axiom that is composed of two or more statements
    """
    __tablename__ = 'owl_complex_axiom'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_complex_axiom(subject={self.subject},predicate={self.predicate},object={self.object},)"
        
    
        
    


class RepairAction(Base):
    """
    Represents an action that needs to be taken to repair a problem
    """
    __tablename__ = 'repair_action'
    
    subject = Column(Text(), primary_key=True)
    description = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"repair_action(subject={self.subject},description={self.description},)"
        
    
        
    


class Problem(Base):
    """
    Represents an instance of a problem pertaining to conformance to OBO guidelines
    """
    __tablename__ = 'problem'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"problem(subject={self.subject},predicate={self.predicate},value={self.value},)"
        
    
        
    


class NodeToNodeStatement(Statements):
    """
    A statement where object is non-null and value is not populated
    """
    __tablename__ = 'node_to_node_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"node_to_node_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class NodeToValueStatement(Statements):
    """
    A statement where value is non-null and object is not populated
    """
    __tablename__ = 'node_to_value_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"node_to_value_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfListStatement(Statements):
    """
    A statement that is used to represent aspects of RDF lists
    """
    __tablename__ = 'rdf_list_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_list_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class BlankNode(Node):
    """
    A node with an ID that is not preserved between databases
    """
    __tablename__ = 'blank_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"blank_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class IriNode(Node):
    """
    
    """
    __tablename__ = 'iri_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"iri_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class ClassNode(Node):
    """
    A node that represents an RDFS/OWL class
    """
    __tablename__ = 'class_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"class_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class PropertyNode(Node):
    """
    Note this only directly classifies nodes asserted to be rdf:Properties
    """
    __tablename__ = 'property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class NamedIndividualNode(Node):
    """
    A node that represents an OWL Named Individual
    """
    __tablename__ = 'named_individual_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"named_individual_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class CountOfPredicates(RdfLevelSummaryStatistic):
    """
    Number of distinct usages of a predicate. NOTE MAY CHANGE: does not currently count existential usage in OWL
    """
    __tablename__ = 'count_of_predicates'
    
    element = Column(Text(), primary_key=True)
    count_value = Column(Integer(), primary_key=True)
    
    
    def __repr__(self):
        return f"count_of_predicates(element={self.element},count_value={self.count_value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class CountOfInstantiatedClasses(RdfLevelSummaryStatistic):
    """
    Number of distinct instantiations of a class. Note in many OBOs, classes are not directly instantiated
    """
    __tablename__ = 'count_of_instantiated_classes'
    
    element = Column(Text(), primary_key=True)
    count_value = Column(Integer(), primary_key=True)
    
    
    def __repr__(self):
        return f"count_of_instantiated_classes(element={self.element},count_value={self.count_value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class CountOfSubclasses(RdfLevelSummaryStatistic):
    """
    Number of subclasses for a given class
    """
    __tablename__ = 'count_of_subclasses'
    
    element = Column(Text(), primary_key=True)
    count_value = Column(Integer(), primary_key=True)
    
    
    def __repr__(self):
        return f"count_of_subclasses(element={self.element},count_value={self.count_value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class Edge(RelationGraphConstruct):
    """
    A relation graph edge that connects two entities by a predicate. Note an edge is distinct from a statement, in that an axiom such as A SubClassOf R some B is represented as multiple statements, but is a single relation graph edge
    """
    __tablename__ = 'edge'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"edge(subject={self.subject},predicate={self.predicate},object={self.object},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SubgraphQuery(RelationGraphConstruct):
    """
    A subgraph query encompasses as subgraph edge and a seed/anchor object and seed/anchor predicate
    """
    __tablename__ = 'subgraph_query'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    anchor_object = Column(Text(), primary_key=True)
    anchor_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subgraph_query(subject={self.subject},predicate={self.predicate},object={self.object},anchor_object={self.anchor_object},anchor_predicate={self.anchor_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class EntailedEdge(RelationGraphConstruct):
    """
    A relation graph edge that is inferred
    """
    __tablename__ = 'entailed_edge'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"entailed_edge(subject={self.subject},predicate={self.predicate},object={self.object},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class EntailedEdgeCycle(RelationGraphConstruct):
    """
    An edge that composes with another edge to make a cycle
    """
    __tablename__ = 'entailed_edge_cycle'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    secondary_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"entailed_edge_cycle(subject={self.subject},predicate={self.predicate},object={self.object},secondary_predicate={self.secondary_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class TransitiveEdge(RelationGraphConstruct):
    """
    A relation graph edge that is formed from a chain of one or more edges
    """
    __tablename__ = 'transitive_edge'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"transitive_edge(subject={self.subject},predicate={self.predicate},object={self.object},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OntologyNode(Node):
    """
    A node representing an ontology
    """
    __tablename__ = 'ontology_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"ontology_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class DeprecatedNode(Node):
    """
    
    """
    __tablename__ = 'deprecated_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"deprecated_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlReifiedAxiom(Statements):
    """
    An OWL axiom that has been reified - i.e. it includes an [id](id) field that uniquely identifies that axiom and which can be the subject of additional statements
    """
    __tablename__ = 'owl_reified_axiom'
    
    id = Column(Text(), primary_key=True)
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_reified_axiom(id={self.id},stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlAxiom(Statements):
    """
    
    """
    __tablename__ = 'owl_axiom'
    
    id = Column(Text(), primary_key=True)
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_axiom(id={self.id},stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlAxiomAnnotation(Statements):
    """
    
    """
    __tablename__ = 'owl_axiom_annotation'
    
    axiom_predicate = Column(Text(), primary_key=True)
    axiom_object = Column(Text(), primary_key=True)
    axiom_value = Column(Text(), primary_key=True)
    axiom_language = Column(Text(), primary_key=True)
    axiom_datatype = Column(Text(), primary_key=True)
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_axiom_annotation(axiom_predicate={self.axiom_predicate},axiom_object={self.axiom_object},axiom_value={self.axiom_value},axiom_language={self.axiom_language},axiom_datatype={self.axiom_datatype},stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlSubclassOfSomeValuesFrom(OwlComplexAxiom):
    """
    Composition of subClassOf and SomeValuesFrom
    """
    __tablename__ = 'owl_subclass_of_some_values_from'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_subclass_of_some_values_from(subject={self.subject},predicate={self.predicate},object={self.object},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlEquivalentToIntersectionMember(OwlComplexAxiom):
    """
    Composition of `OwlEquivalentClass`, `OwlIntersectionOf`, and `RdfListMember`; `C = X1 and ... and Xn`
    """
    __tablename__ = 'owl_equivalent_to_intersection_member'
    
    subject = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_equivalent_to_intersection_member(subject={self.subject},object={self.object},predicate={self.predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class LexicalProblem(Problem):
    """
    a problem with the textual value of an annotation property
    """
    __tablename__ = 'lexical_problem'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"lexical_problem(subject={self.subject},predicate={self.predicate},value={self.value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class TrailingWhitespaceProblem(Problem):
    """
    
    """
    __tablename__ = 'trailing_whitespace_problem'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"trailing_whitespace_problem(subject={self.subject},predicate={self.predicate},value={self.value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class PropertyUsedWithDatatypeValuesAndObjects(Problem):
    """
    A problem in which the same property is used two two different ways, one in which the range is a literal value, the other where it is an object.
    """
    __tablename__ = 'property_used_with_datatype_values_and_objects'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"property_used_with_datatype_values_and_objects(subject={self.subject},predicate={self.predicate},value={self.value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class NodeWithTwoLabelsProblem(Problem):
    """
    
    """
    __tablename__ = 'node_with_two_labels_problem'
    
    label1 = Column(Text(), primary_key=True)
    label2 = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"node_with_two_labels_problem(label1={self.label1},label2={self.label2},subject={self.subject},predicate={self.predicate},value={self.value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AllProblems(Problem):
    """
    
    """
    __tablename__ = 'all_problems'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"all_problems(subject={self.subject},predicate={self.predicate},value={self.value},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfTypeStatement(NodeToNodeStatement):
    """
    A statement that indicates the asserted type of the subject entity
    """
    __tablename__ = 'rdf_type_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_type_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfsSubclassOfStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'rdfs_subclass_of_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdfs_subclass_of_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfsSubpropertyOfStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'rdfs_subproperty_of_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdfs_subproperty_of_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfsLabelStatement(NodeToValueStatement):
    """
    
    """
    __tablename__ = 'rdfs_label_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdfs_label_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfsDomainStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'rdfs_domain_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdfs_domain_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfsRangeStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'rdfs_range_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdfs_range_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfFirstStatement(RdfListStatement):
    """
    A statement that connects a list to its first element. This is a low-level triple, it is unlikely you need to use this directly. It is used to define rdf_list_member_statement, which is more useful
    """
    __tablename__ = 'rdf_first_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_first_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfRestStatement(RdfListStatement):
    """
    A statement that connects a list to its remaining elements. This is a low-level triple, it is unlikely you need to use this directly. It is used to define rdf_list_member_statement, which is more useful
    """
    __tablename__ = 'rdf_rest_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_rest_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfRestTransitiveStatement(RdfListStatement):
    """
    
    """
    __tablename__ = 'rdf_rest_transitive_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_rest_transitive_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfListMemberStatement(RdfListStatement):
    """
    
    """
    __tablename__ = 'rdf_list_member_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_list_member_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfListNode(BlankNode):
    """
    A node representing an RDF list. Note that you will not likely need to use this directly.
    """
    __tablename__ = 'rdf_list_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdf_list_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SubgraphEdgeByAncestor(SubgraphQuery):
    """
    An edge within a subgraph anchored around a set of ancestor terms
    """
    __tablename__ = 'subgraph_edge_by_ancestor'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    anchor_object = Column(Text(), primary_key=True)
    anchor_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subgraph_edge_by_ancestor(subject={self.subject},predicate={self.predicate},object={self.object},anchor_object={self.anchor_object},anchor_predicate={self.anchor_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SubgraphEdgeByDescendant(SubgraphQuery):
    """
    An edge within a subgraph anchored around a set of descendant terms
    """
    __tablename__ = 'subgraph_edge_by_descendant'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    anchor_object = Column(Text(), primary_key=True)
    anchor_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subgraph_edge_by_descendant(subject={self.subject},predicate={self.predicate},object={self.object},anchor_object={self.anchor_object},anchor_predicate={self.anchor_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SubgraphEdgeByAncestorOrDescendant(SubgraphQuery):
    """
    
    """
    __tablename__ = 'subgraph_edge_by_ancestor_or_descendant'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    anchor_object = Column(Text(), primary_key=True)
    anchor_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subgraph_edge_by_ancestor_or_descendant(subject={self.subject},predicate={self.predicate},object={self.object},anchor_object={self.anchor_object},anchor_predicate={self.anchor_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SubgraphEdgeByParent(SubgraphQuery):
    """
    An edge within a subgraph anchored around a set of parent terms
    """
    __tablename__ = 'subgraph_edge_by_parent'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    anchor_object = Column(Text(), primary_key=True)
    anchor_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subgraph_edge_by_parent(subject={self.subject},predicate={self.predicate},object={self.object},anchor_object={self.anchor_object},anchor_predicate={self.anchor_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SubgraphEdgeByChild(SubgraphQuery):
    """
    An edge within a subgraph anchored around a set of child terms
    """
    __tablename__ = 'subgraph_edge_by_child'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    anchor_object = Column(Text(), primary_key=True)
    anchor_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subgraph_edge_by_child(subject={self.subject},predicate={self.predicate},object={self.object},anchor_object={self.anchor_object},anchor_predicate={self.anchor_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SubgraphEdgeBySelf(SubgraphQuery):
    """
    A special null form of a subgraph query where there is no expansion
    """
    __tablename__ = 'subgraph_edge_by_self'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    anchor_object = Column(Text(), primary_key=True)
    anchor_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"subgraph_edge_by_self(subject={self.subject},predicate={self.predicate},object={self.object},anchor_object={self.anchor_object},anchor_predicate={self.anchor_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class EntailedEdgeSamePredicateCycle(EntailedEdgeCycle):
    """
    An entailed_edge_cycle over a single predicate
    """
    __tablename__ = 'entailed_edge_same_predicate_cycle'
    
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    secondary_predicate = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"entailed_edge_same_predicate_cycle(subject={self.subject},predicate={self.predicate},object={self.object},secondary_predicate={self.secondary_predicate},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class ObjectPropertyNode(PropertyNode):
    """
    A node representing an OWL object property
    """
    __tablename__ = 'object_property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"object_property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AnnotationPropertyNode(PropertyNode):
    """
    A node representing an OWL annotation property
    """
    __tablename__ = 'annotation_property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"annotation_property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlImportsStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'owl_imports_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_imports_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlInverseOfStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'owl_inverse_of_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_inverse_of_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlComplementOfStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'owl_complement_of_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_complement_of_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlEquivalentClassStatement(NodeToNodeStatement):
    """
    A statement that connects two class_nodes where both classes are equivalent
    """
    __tablename__ = 'owl_equivalent_class_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_equivalent_class_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlSameAsStatement(NodeToNodeStatement):
    """
    A statement that connects two individual nodes where both individual are equivalent
    """
    __tablename__ = 'owl_same_as_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_same_as_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlDisjointClassStatement(NodeToNodeStatement):
    """
    
    """
    __tablename__ = 'owl_disjoint_class_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_disjoint_class_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AnonymousExpression(BlankNode):
    """
    An OWL expression, such as a class expression. Expressions are "anonymous" as they are a composition of named elements rather than a named element themselves
    """
    __tablename__ = 'anonymous_expression'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"anonymous_expression(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasTextDefinitionStatement(NodeToValueStatement):
    """
    
    """
    __tablename__ = 'has_text_definition_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_text_definition_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasOioSynonymStatement(NodeToValueStatement):
    """
    
    """
    __tablename__ = 'has_oio_synonym_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_oio_synonym_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasSynonymStatement(NodeToValueStatement):
    """
    
    """
    __tablename__ = 'has_synonym_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_synonym_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasMatchStatement(NodeToValueStatement):
    """
    
    """
    __tablename__ = 'has_match_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_match_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasMappingStatement(NodeToValueStatement):
    """
    
    """
    __tablename__ = 'has_mapping_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_mapping_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AxiomDbxrefAnnotation(OwlAxiomAnnotation):
    """
    
    """
    __tablename__ = 'axiom_dbxref_annotation'
    
    axiom_predicate = Column(Text(), primary_key=True)
    axiom_object = Column(Text(), primary_key=True)
    axiom_value = Column(Text(), primary_key=True)
    axiom_language = Column(Text(), primary_key=True)
    axiom_datatype = Column(Text(), primary_key=True)
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"axiom_dbxref_annotation(axiom_predicate={self.axiom_predicate},axiom_object={self.axiom_object},axiom_value={self.axiom_value},axiom_language={self.axiom_language},axiom_datatype={self.axiom_datatype},stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OntologyStatusStatement(NodeToValueStatement):
    """
    
    """
    __tablename__ = 'ontology_status_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"ontology_status_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class RdfsSubclassOfNamedStatement(RdfsSubclassOfStatement):
    """
    
    """
    __tablename__ = 'rdfs_subclass_of_named_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"rdfs_subclass_of_named_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class TransitivePropertyNode(ObjectPropertyNode):
    """
    A node representing an OWL transitive object property
    """
    __tablename__ = 'transitive_property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"transitive_property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class SymmetricPropertyNode(ObjectPropertyNode):
    """
    A node representing an OWL symmetric object property
    """
    __tablename__ = 'symmetric_property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"symmetric_property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class ReflexivePropertyNode(ObjectPropertyNode):
    """
    A node representing an OWL reflexive object property
    """
    __tablename__ = 'reflexive_property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"reflexive_property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class IrreflexivePropertyNode(ObjectPropertyNode):
    """
    A node representing an OWL irreflexive object property
    """
    __tablename__ = 'irreflexive_property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"irreflexive_property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AsymmetricPropertyNode(ObjectPropertyNode):
    """
    
    """
    __tablename__ = 'asymmetric_property_node'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"asymmetric_property_node(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AnonymousClassExpression(AnonymousExpression):
    """
    An OWL anonymous class expression, such as for example `SomeValuesFrom(partOf Hand)`
    """
    __tablename__ = 'anonymous_class_expression'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"anonymous_class_expression(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AnonymousPropertyExpression(AnonymousExpression):
    """
    
    """
    __tablename__ = 'anonymous_property_expression'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"anonymous_property_expression(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class AnonymousIndividualExpression(AnonymousExpression):
    """
    
    """
    __tablename__ = 'anonymous_individual_expression'
    
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"anonymous_individual_expression(id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasExactSynonymStatement(HasOioSynonymStatement):
    """
    
    """
    __tablename__ = 'has_exact_synonym_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_exact_synonym_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasBroadSynonymStatement(HasOioSynonymStatement):
    """
    
    """
    __tablename__ = 'has_broad_synonym_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_broad_synonym_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasNarrowSynonymStatement(HasOioSynonymStatement):
    """
    
    """
    __tablename__ = 'has_narrow_synonym_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_narrow_synonym_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasRelatedSynonymStatement(HasOioSynonymStatement):
    """
    
    """
    __tablename__ = 'has_related_synonym_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_related_synonym_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasExactMatchStatement(HasMatchStatement):
    """
    
    """
    __tablename__ = 'has_exact_match_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_exact_match_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasBroadMatchStatement(HasMatchStatement):
    """
    
    """
    __tablename__ = 'has_broad_match_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_broad_match_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasNarrowMatchStatement(HasMatchStatement):
    """
    
    """
    __tablename__ = 'has_narrow_match_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_narrow_match_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasRelatedMatchStatement(HasMatchStatement):
    """
    
    """
    __tablename__ = 'has_related_match_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_related_match_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class HasDbxrefStatement(HasMappingStatement):
    """
    
    """
    __tablename__ = 'has_dbxref_statement'
    
    stanza = Column(Text(), primary_key=True)
    subject = Column(Text(), primary_key=True)
    predicate = Column(Text(), primary_key=True)
    object = Column(Text(), primary_key=True)
    value = Column(Text(), primary_key=True)
    datatype = Column(Text(), primary_key=True)
    language = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"has_dbxref_statement(stanza={self.stanza},subject={self.subject},predicate={self.predicate},object={self.object},value={self.value},datatype={self.datatype},language={self.language},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlRestriction(AnonymousClassExpression):
    """
    An OWL restriction, such as `SomeValuesFrom(partOf Hand)`
    """
    __tablename__ = 'owl_restriction'
    
    on_property = Column(Text(), primary_key=True)
    filler = Column(Text(), primary_key=True)
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_restriction(on_property={self.on_property},filler={self.filler},id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlSomeValuesFrom(OwlRestriction):
    """
    An OWL SomeValuesFrom restriction
    """
    __tablename__ = 'owl_some_values_from'
    
    on_property = Column(Text(), primary_key=True)
    filler = Column(Text(), primary_key=True)
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_some_values_from(on_property={self.on_property},filler={self.filler},id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlAllValuesFrom(OwlRestriction):
    """
    
    """
    __tablename__ = 'owl_all_values_from'
    
    on_property = Column(Text(), primary_key=True)
    filler = Column(Text(), primary_key=True)
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_all_values_from(on_property={self.on_property},filler={self.filler},id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlHasValue(OwlRestriction):
    """
    
    """
    __tablename__ = 'owl_has_value'
    
    on_property = Column(Text(), primary_key=True)
    filler = Column(Text(), primary_key=True)
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_has_value(on_property={self.on_property},filler={self.filler},id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    


class OwlHasSelf(OwlRestriction):
    """
    
    """
    __tablename__ = 'owl_has_self'
    
    on_property = Column(Text(), primary_key=True)
    filler = Column(Text(), primary_key=True)
    id = Column(Text(), primary_key=True)
    
    
    def __repr__(self):
        return f"owl_has_self(on_property={self.on_property},filler={self.filler},id={self.id},)"
        
    
        
    
    # Using concrete inheritance: see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    __mapper_args__ = {
        'concrete': True
    }
    





# Auto generated from summary_statistics_datamodel.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-03-29T19:03:43
# Schema: summary-statistics
#
# id: https://w3id.org/linkml/summary_statistics
# description: A datamodel for reports on data
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Integer, String

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
REPORTING = CurieNamespace('reporting', 'https://w3id.org/linkml/report')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SH = CurieNamespace('sh', 'https://w3id.org/shacl/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = REPORTING


# Types

# Class references
class FacetStatisticsFacet(extended_str):
    pass


@dataclass
class SummaryStatisticCollection(YAMLRoot):
    """
    A summary statistics report object
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.SummaryStatisticCollection
    class_class_curie: ClassVar[str] = "reporting:SummaryStatisticCollection"
    class_name: ClassVar[str] = "SummaryStatisticCollection"
    class_model_uri: ClassVar[URIRef] = REPORTING.SummaryStatisticCollection

    class_count: Optional[int] = None
    anonymous_class_expression_count: Optional[int] = None
    unsatisfiable_class_count: Optional[int] = None
    class_count_excluding_deprecated: Optional[int] = None
    class_count_with_definitions: Optional[int] = None
    property_count: Optional[int] = None
    object_property_count: Optional[int] = None
    datatype_property_count: Optional[int] = None
    annotation_property_count: Optional[int] = None
    individual_count: Optional[int] = None
    named_individual_count: Optional[int] = None
    anonymous_individual_count: Optional[int] = None
    untyped_entity_count: Optional[int] = None
    description_logic_profile: Optional[str] = None
    owl_axiom_count: Optional[int] = None
    rdf_triple_count: Optional[int] = None
    subclass_of_axiom_count: Optional[int] = None
    equivalentclasses_axiom_count: Optional[int] = None
    distinct_synonym_count: Optional[int] = None
    synonym_statement_count: Optional[int] = None
    mapping_count: Optional[int] = None
    ontology_count: Optional[int] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.class_count is not None and not isinstance(self.class_count, int):
            self.class_count = int(self.class_count)

        if self.anonymous_class_expression_count is not None and not isinstance(self.anonymous_class_expression_count, int):
            self.anonymous_class_expression_count = int(self.anonymous_class_expression_count)

        if self.unsatisfiable_class_count is not None and not isinstance(self.unsatisfiable_class_count, int):
            self.unsatisfiable_class_count = int(self.unsatisfiable_class_count)

        if self.class_count_excluding_deprecated is not None and not isinstance(self.class_count_excluding_deprecated, int):
            self.class_count_excluding_deprecated = int(self.class_count_excluding_deprecated)

        if self.class_count_with_definitions is not None and not isinstance(self.class_count_with_definitions, int):
            self.class_count_with_definitions = int(self.class_count_with_definitions)

        if self.property_count is not None and not isinstance(self.property_count, int):
            self.property_count = int(self.property_count)

        if self.object_property_count is not None and not isinstance(self.object_property_count, int):
            self.object_property_count = int(self.object_property_count)

        if self.datatype_property_count is not None and not isinstance(self.datatype_property_count, int):
            self.datatype_property_count = int(self.datatype_property_count)

        if self.annotation_property_count is not None and not isinstance(self.annotation_property_count, int):
            self.annotation_property_count = int(self.annotation_property_count)

        if self.individual_count is not None and not isinstance(self.individual_count, int):
            self.individual_count = int(self.individual_count)

        if self.named_individual_count is not None and not isinstance(self.named_individual_count, int):
            self.named_individual_count = int(self.named_individual_count)

        if self.anonymous_individual_count is not None and not isinstance(self.anonymous_individual_count, int):
            self.anonymous_individual_count = int(self.anonymous_individual_count)

        if self.untyped_entity_count is not None and not isinstance(self.untyped_entity_count, int):
            self.untyped_entity_count = int(self.untyped_entity_count)

        if self.description_logic_profile is not None and not isinstance(self.description_logic_profile, str):
            self.description_logic_profile = str(self.description_logic_profile)

        if self.owl_axiom_count is not None and not isinstance(self.owl_axiom_count, int):
            self.owl_axiom_count = int(self.owl_axiom_count)

        if self.rdf_triple_count is not None and not isinstance(self.rdf_triple_count, int):
            self.rdf_triple_count = int(self.rdf_triple_count)

        if self.subclass_of_axiom_count is not None and not isinstance(self.subclass_of_axiom_count, int):
            self.subclass_of_axiom_count = int(self.subclass_of_axiom_count)

        if self.equivalentclasses_axiom_count is not None and not isinstance(self.equivalentclasses_axiom_count, int):
            self.equivalentclasses_axiom_count = int(self.equivalentclasses_axiom_count)

        if self.distinct_synonym_count is not None and not isinstance(self.distinct_synonym_count, int):
            self.distinct_synonym_count = int(self.distinct_synonym_count)

        if self.synonym_statement_count is not None and not isinstance(self.synonym_statement_count, int):
            self.synonym_statement_count = int(self.synonym_statement_count)

        if self.mapping_count is not None and not isinstance(self.mapping_count, int):
            self.mapping_count = int(self.mapping_count)

        if self.ontology_count is not None and not isinstance(self.ontology_count, int):
            self.ontology_count = int(self.ontology_count)

        super().__post_init__(**kwargs)


@dataclass
class GlobalStatistics(SummaryStatisticCollection):
    """
    summary statistics for the entire resource
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.GlobalStatistics
    class_class_curie: ClassVar[str] = "reporting:GlobalStatistics"
    class_name: ClassVar[str] = "GlobalStatistics"
    class_model_uri: ClassVar[URIRef] = REPORTING.GlobalStatistics

    by_category: Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, "FacetStatistics"]], List[Union[dict, "FacetStatistics"]]]] = empty_dict()
    by_taxon: Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, "FacetStatistics"]], List[Union[dict, "FacetStatistics"]]]] = empty_dict()
    by_ontology: Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, "FacetStatistics"]], List[Union[dict, "FacetStatistics"]]]] = empty_dict()
    by_subset: Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, "FacetStatistics"]], List[Union[dict, "FacetStatistics"]]]] = empty_dict()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        self._normalize_inlined_as_dict(slot_name="by_category", slot_type=FacetStatistics, key_name="facet", keyed=True)

        self._normalize_inlined_as_dict(slot_name="by_taxon", slot_type=FacetStatistics, key_name="facet", keyed=True)

        self._normalize_inlined_as_dict(slot_name="by_ontology", slot_type=FacetStatistics, key_name="facet", keyed=True)

        self._normalize_inlined_as_dict(slot_name="by_subset", slot_type=FacetStatistics, key_name="facet", keyed=True)

        super().__post_init__(**kwargs)


@dataclass
class FacetStatistics(SummaryStatisticCollection):
    """
    summary statistics for a data facet
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.FacetStatistics
    class_class_curie: ClassVar[str] = "reporting:FacetStatistics"
    class_name: ClassVar[str] = "FacetStatistics"
    class_model_uri: ClassVar[URIRef] = REPORTING.FacetStatistics

    facet: Union[str, FacetStatisticsFacet] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.facet):
            self.MissingRequiredField("facet")
        if not isinstance(self.facet, FacetStatisticsFacet):
            self.facet = FacetStatisticsFacet(self.facet)

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass

slots.count_statistic = Slot(uri=REPORTING.count_statistic, name="count_statistic", curie=REPORTING.curie('count_statistic'),
                   model_uri=REPORTING.count_statistic, domain=None, range=Optional[int])

slots.class_statistic_group = Slot(uri=REPORTING.class_statistic_group, name="class_statistic_group", curie=REPORTING.curie('class_statistic_group'),
                   model_uri=REPORTING.class_statistic_group, domain=None, range=Optional[str])

slots.property_statistic_group = Slot(uri=REPORTING.property_statistic_group, name="property_statistic_group", curie=REPORTING.curie('property_statistic_group'),
                   model_uri=REPORTING.property_statistic_group, domain=None, range=Optional[str])

slots.individual_statistic_group = Slot(uri=REPORTING.individual_statistic_group, name="individual_statistic_group", curie=REPORTING.curie('individual_statistic_group'),
                   model_uri=REPORTING.individual_statistic_group, domain=None, range=Optional[str])

slots.metadata_statistic_group = Slot(uri=REPORTING.metadata_statistic_group, name="metadata_statistic_group", curie=REPORTING.curie('metadata_statistic_group'),
                   model_uri=REPORTING.metadata_statistic_group, domain=None, range=Optional[str])

slots.owl_statistic_group = Slot(uri=REPORTING.owl_statistic_group, name="owl_statistic_group", curie=REPORTING.curie('owl_statistic_group'),
                   model_uri=REPORTING.owl_statistic_group, domain=None, range=Optional[str])

slots.summaryStatisticCollection__class_count = Slot(uri=REPORTING.class_count, name="summaryStatisticCollection__class_count", curie=REPORTING.curie('class_count'),
                   model_uri=REPORTING.summaryStatisticCollection__class_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__anonymous_class_expression_count = Slot(uri=REPORTING.anonymous_class_expression_count, name="summaryStatisticCollection__anonymous_class_expression_count", curie=REPORTING.curie('anonymous_class_expression_count'),
                   model_uri=REPORTING.summaryStatisticCollection__anonymous_class_expression_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__unsatisfiable_class_count = Slot(uri=REPORTING.unsatisfiable_class_count, name="summaryStatisticCollection__unsatisfiable_class_count", curie=REPORTING.curie('unsatisfiable_class_count'),
                   model_uri=REPORTING.summaryStatisticCollection__unsatisfiable_class_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__class_count_excluding_deprecated = Slot(uri=REPORTING.class_count_excluding_deprecated, name="summaryStatisticCollection__class_count_excluding_deprecated", curie=REPORTING.curie('class_count_excluding_deprecated'),
                   model_uri=REPORTING.summaryStatisticCollection__class_count_excluding_deprecated, domain=None, range=Optional[int])

slots.summaryStatisticCollection__class_count_with_definitions = Slot(uri=REPORTING.class_count_with_definitions, name="summaryStatisticCollection__class_count_with_definitions", curie=REPORTING.curie('class_count_with_definitions'),
                   model_uri=REPORTING.summaryStatisticCollection__class_count_with_definitions, domain=None, range=Optional[int])

slots.summaryStatisticCollection__property_count = Slot(uri=REPORTING.property_count, name="summaryStatisticCollection__property_count", curie=REPORTING.curie('property_count'),
                   model_uri=REPORTING.summaryStatisticCollection__property_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__object_property_count = Slot(uri=REPORTING.object_property_count, name="summaryStatisticCollection__object_property_count", curie=REPORTING.curie('object_property_count'),
                   model_uri=REPORTING.summaryStatisticCollection__object_property_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__datatype_property_count = Slot(uri=REPORTING.datatype_property_count, name="summaryStatisticCollection__datatype_property_count", curie=REPORTING.curie('datatype_property_count'),
                   model_uri=REPORTING.summaryStatisticCollection__datatype_property_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__annotation_property_count = Slot(uri=REPORTING.annotation_property_count, name="summaryStatisticCollection__annotation_property_count", curie=REPORTING.curie('annotation_property_count'),
                   model_uri=REPORTING.summaryStatisticCollection__annotation_property_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__individual_count = Slot(uri=REPORTING.individual_count, name="summaryStatisticCollection__individual_count", curie=REPORTING.curie('individual_count'),
                   model_uri=REPORTING.summaryStatisticCollection__individual_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__named_individual_count = Slot(uri=REPORTING.named_individual_count, name="summaryStatisticCollection__named_individual_count", curie=REPORTING.curie('named_individual_count'),
                   model_uri=REPORTING.summaryStatisticCollection__named_individual_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__anonymous_individual_count = Slot(uri=REPORTING.anonymous_individual_count, name="summaryStatisticCollection__anonymous_individual_count", curie=REPORTING.curie('anonymous_individual_count'),
                   model_uri=REPORTING.summaryStatisticCollection__anonymous_individual_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__untyped_entity_count = Slot(uri=REPORTING.untyped_entity_count, name="summaryStatisticCollection__untyped_entity_count", curie=REPORTING.curie('untyped_entity_count'),
                   model_uri=REPORTING.summaryStatisticCollection__untyped_entity_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__description_logic_profile = Slot(uri=REPORTING.description_logic_profile, name="summaryStatisticCollection__description_logic_profile", curie=REPORTING.curie('description_logic_profile'),
                   model_uri=REPORTING.summaryStatisticCollection__description_logic_profile, domain=None, range=Optional[str])

slots.summaryStatisticCollection__owl_axiom_count = Slot(uri=REPORTING.owl_axiom_count, name="summaryStatisticCollection__owl_axiom_count", curie=REPORTING.curie('owl_axiom_count'),
                   model_uri=REPORTING.summaryStatisticCollection__owl_axiom_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__rdf_triple_count = Slot(uri=REPORTING.rdf_triple_count, name="summaryStatisticCollection__rdf_triple_count", curie=REPORTING.curie('rdf_triple_count'),
                   model_uri=REPORTING.summaryStatisticCollection__rdf_triple_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__subclass_of_axiom_count = Slot(uri=REPORTING.subclass_of_axiom_count, name="summaryStatisticCollection__subclass_of_axiom_count", curie=REPORTING.curie('subclass_of_axiom_count'),
                   model_uri=REPORTING.summaryStatisticCollection__subclass_of_axiom_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__equivalentclasses_axiom_count = Slot(uri=REPORTING.equivalentclasses_axiom_count, name="summaryStatisticCollection__equivalentclasses_axiom_count", curie=REPORTING.curie('equivalentclasses_axiom_count'),
                   model_uri=REPORTING.summaryStatisticCollection__equivalentclasses_axiom_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__distinct_synonym_count = Slot(uri=REPORTING.distinct_synonym_count, name="summaryStatisticCollection__distinct_synonym_count", curie=REPORTING.curie('distinct_synonym_count'),
                   model_uri=REPORTING.summaryStatisticCollection__distinct_synonym_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__synonym_statement_count = Slot(uri=REPORTING.synonym_statement_count, name="summaryStatisticCollection__synonym_statement_count", curie=REPORTING.curie('synonym_statement_count'),
                   model_uri=REPORTING.summaryStatisticCollection__synonym_statement_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__mapping_count = Slot(uri=REPORTING.mapping_count, name="summaryStatisticCollection__mapping_count", curie=REPORTING.curie('mapping_count'),
                   model_uri=REPORTING.summaryStatisticCollection__mapping_count, domain=None, range=Optional[int])

slots.summaryStatisticCollection__ontology_count = Slot(uri=REPORTING.ontology_count, name="summaryStatisticCollection__ontology_count", curie=REPORTING.curie('ontology_count'),
                   model_uri=REPORTING.summaryStatisticCollection__ontology_count, domain=None, range=Optional[int])

slots.globalStatistics__by_category = Slot(uri=REPORTING.by_category, name="globalStatistics__by_category", curie=REPORTING.curie('by_category'),
                   model_uri=REPORTING.globalStatistics__by_category, domain=None, range=Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, FacetStatistics]], List[Union[dict, FacetStatistics]]]])

slots.globalStatistics__by_taxon = Slot(uri=REPORTING.by_taxon, name="globalStatistics__by_taxon", curie=REPORTING.curie('by_taxon'),
                   model_uri=REPORTING.globalStatistics__by_taxon, domain=None, range=Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, FacetStatistics]], List[Union[dict, FacetStatistics]]]])

slots.globalStatistics__by_ontology = Slot(uri=REPORTING.by_ontology, name="globalStatistics__by_ontology", curie=REPORTING.curie('by_ontology'),
                   model_uri=REPORTING.globalStatistics__by_ontology, domain=None, range=Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, FacetStatistics]], List[Union[dict, FacetStatistics]]]])

slots.globalStatistics__by_subset = Slot(uri=REPORTING.by_subset, name="globalStatistics__by_subset", curie=REPORTING.curie('by_subset'),
                   model_uri=REPORTING.globalStatistics__by_subset, domain=None, range=Optional[Union[Dict[Union[str, FacetStatisticsFacet], Union[dict, FacetStatistics]], List[Union[dict, FacetStatistics]]]])

slots.facetStatistics__facet = Slot(uri=REPORTING.facet, name="facetStatistics__facet", curie=REPORTING.curie('facet'),
                   model_uri=REPORTING.facetStatistics__facet, domain=None, range=URIRef)



# Auto generated from mapping_rules_datamodel.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-03-31T13:26:26
# Schema: mapping-rules
#
# id: https://w3id.org/linkml/mapping_rules_datamodel
# description: A datamodel for specifying lexical mapping rules. NOTE -- this may move to another package
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, Float, String, Uriorcurie
from linkml_runtime.utils.metamodelcore import Bool, URIorCURIE

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LI = CurieNamespace('li', 'https://w3id.org/linkml/lexical_index/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
MRULES = CurieNamespace('mrules', 'https://w3id.org/linkml/mapping_rules_datamodel/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
PROV = CurieNamespace('prov', 'http://www.w3.org/ns/prov#')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SH = CurieNamespace('sh', 'https://w3id.org/shacl/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = MRULES


# Types

# Class references
class LexicalGroupingTerm(extended_str):
    pass


class LexicalTransformationPipelineName(extended_str):
    pass


@dataclass
class MappingRuleCollection(YAMLRoot):
    """
    A collection of mapping rules
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = MRULES.MappingRuleCollection
    class_class_curie: ClassVar[str] = "mrules:MappingRuleCollection"
    class_name: ClassVar[str] = "MappingRuleCollection"
    class_model_uri: ClassVar[URIRef] = MRULES.MappingRuleCollection

    rules: Optional[Union[Union[dict, "MappingRule"], List[Union[dict, "MappingRule"]]]] = empty_list()
    minimum_confidence: Optional[float] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.rules, list):
            self.rules = [self.rules] if self.rules is not None else []
        self.rules = [v if isinstance(v, MappingRule) else MappingRule(**as_dict(v)) for v in self.rules]

        if self.minimum_confidence is not None and not isinstance(self.minimum_confidence, float):
            self.minimum_confidence = float(self.minimum_confidence)

        super().__post_init__(**kwargs)


@dataclass
class MappingRule(YAMLRoot):
    """
    An individual mapping rule, if preconditions match the postconditions are applied
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = MRULES.MappingRule
    class_class_curie: ClassVar[str] = "mrules:MappingRule"
    class_name: ClassVar[str] = "MappingRule"
    class_model_uri: ClassVar[URIRef] = MRULES.MappingRule

    description: Optional[str] = None
    oneway: Optional[Union[bool, Bool]] = False
    preconditions: Optional[Union[dict, "Precondition"]] = None
    postconditions: Optional[Union[dict, "Postcondition"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.oneway is not None and not isinstance(self.oneway, Bool):
            self.oneway = Bool(self.oneway)

        if self.preconditions is not None and not isinstance(self.preconditions, Precondition):
            self.preconditions = Precondition(**as_dict(self.preconditions))

        if self.postconditions is not None and not isinstance(self.postconditions, Postcondition):
            self.postconditions = Postcondition(**as_dict(self.postconditions))

        super().__post_init__(**kwargs)


@dataclass
class Precondition(YAMLRoot):
    """
    A pattern to be matched against an individual SSSOM mapping
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = MRULES.Precondition
    class_class_curie: ClassVar[str] = "mrules:Precondition"
    class_name: ClassVar[str] = "Precondition"
    class_model_uri: ClassVar[URIRef] = MRULES.Precondition

    subject_source_one_of: Optional[Union[str, List[str]]] = empty_list()
    object_source_one_of: Optional[Union[str, List[str]]] = empty_list()
    mapping_source_one_of: Optional[Union[str, List[str]]] = empty_list()
    subject_match_field_one_of: Optional[Union[str, List[str]]] = empty_list()
    object_match_field_one_of: Optional[Union[str, List[str]]] = empty_list()
    transformations_included_in: Optional[Union[str, List[str]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.subject_source_one_of, list):
            self.subject_source_one_of = [self.subject_source_one_of] if self.subject_source_one_of is not None else []
        self.subject_source_one_of = [v if isinstance(v, str) else str(v) for v in self.subject_source_one_of]

        if not isinstance(self.object_source_one_of, list):
            self.object_source_one_of = [self.object_source_one_of] if self.object_source_one_of is not None else []
        self.object_source_one_of = [v if isinstance(v, str) else str(v) for v in self.object_source_one_of]

        if not isinstance(self.mapping_source_one_of, list):
            self.mapping_source_one_of = [self.mapping_source_one_of] if self.mapping_source_one_of is not None else []
        self.mapping_source_one_of = [v if isinstance(v, str) else str(v) for v in self.mapping_source_one_of]

        if not isinstance(self.subject_match_field_one_of, list):
            self.subject_match_field_one_of = [self.subject_match_field_one_of] if self.subject_match_field_one_of is not None else []
        self.subject_match_field_one_of = [v if isinstance(v, str) else str(v) for v in self.subject_match_field_one_of]

        if not isinstance(self.object_match_field_one_of, list):
            self.object_match_field_one_of = [self.object_match_field_one_of] if self.object_match_field_one_of is not None else []
        self.object_match_field_one_of = [v if isinstance(v, str) else str(v) for v in self.object_match_field_one_of]

        if not isinstance(self.transformations_included_in, list):
            self.transformations_included_in = [self.transformations_included_in] if self.transformations_included_in is not None else []
        self.transformations_included_in = [v if isinstance(v, str) else str(v) for v in self.transformations_included_in]

        super().__post_init__(**kwargs)


@dataclass
class Postcondition(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = MRULES.Postcondition
    class_class_curie: ClassVar[str] = "mrules:Postcondition"
    class_name: ClassVar[str] = "Postcondition"
    class_model_uri: ClassVar[URIRef] = MRULES.Postcondition

    predicate_id: Optional[str] = None
    weight: Optional[float] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.predicate_id is not None and not isinstance(self.predicate_id, str):
            self.predicate_id = str(self.predicate_id)

        if self.weight is not None and not isinstance(self.weight, float):
            self.weight = float(self.weight)

        super().__post_init__(**kwargs)


@dataclass
class LexicalIndex(YAMLRoot):
    """
    An index over an ontology keyed by lexical unit
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LI.LexicalIndex
    class_class_curie: ClassVar[str] = "li:LexicalIndex"
    class_name: ClassVar[str] = "LexicalIndex"
    class_model_uri: ClassVar[URIRef] = MRULES.LexicalIndex

    groupings: Optional[Union[Dict[Union[str, LexicalGroupingTerm], Union[dict, "LexicalGrouping"]], List[Union[dict, "LexicalGrouping"]]]] = empty_dict()
    pipelines: Optional[Union[Dict[Union[str, LexicalTransformationPipelineName], Union[dict, "LexicalTransformationPipeline"]], List[Union[dict, "LexicalTransformationPipeline"]]]] = empty_dict()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        self._normalize_inlined_as_dict(slot_name="groupings", slot_type=LexicalGrouping, key_name="term", keyed=True)

        self._normalize_inlined_as_dict(slot_name="pipelines", slot_type=LexicalTransformationPipeline, key_name="name", keyed=True)

        super().__post_init__(**kwargs)


@dataclass
class LexicalGrouping(YAMLRoot):
    """
    A grouping of ontology elements by a shared lexical term
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LI.LexicalGrouping
    class_class_curie: ClassVar[str] = "li:LexicalGrouping"
    class_name: ClassVar[str] = "LexicalGrouping"
    class_model_uri: ClassVar[URIRef] = MRULES.LexicalGrouping

    term: Union[str, LexicalGroupingTerm] = None
    relationships: Optional[Union[Union[dict, "RelationshipToTerm"], List[Union[dict, "RelationshipToTerm"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.term):
            self.MissingRequiredField("term")
        if not isinstance(self.term, LexicalGroupingTerm):
            self.term = LexicalGroupingTerm(self.term)

        if not isinstance(self.relationships, list):
            self.relationships = [self.relationships] if self.relationships is not None else []
        self.relationships = [v if isinstance(v, RelationshipToTerm) else RelationshipToTerm(**as_dict(v)) for v in self.relationships]

        super().__post_init__(**kwargs)


@dataclass
class RelationshipToTerm(YAMLRoot):
    """
    A relationship of an ontology element to a lexical term
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LI.RelationshipToTerm
    class_class_curie: ClassVar[str] = "li:RelationshipToTerm"
    class_name: ClassVar[str] = "RelationshipToTerm"
    class_model_uri: ClassVar[URIRef] = MRULES.RelationshipToTerm

    predicate: Optional[Union[str, URIorCURIE]] = None
    element: Optional[Union[str, URIorCURIE]] = None
    element_term: Optional[str] = None
    source: Optional[Union[str, URIorCURIE]] = None
    pipeline: Optional[Union[Union[str, LexicalTransformationPipelineName], List[Union[str, LexicalTransformationPipelineName]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.predicate is not None and not isinstance(self.predicate, URIorCURIE):
            self.predicate = URIorCURIE(self.predicate)

        if self.element is not None and not isinstance(self.element, URIorCURIE):
            self.element = URIorCURIE(self.element)

        if self.element_term is not None and not isinstance(self.element_term, str):
            self.element_term = str(self.element_term)

        if self.source is not None and not isinstance(self.source, URIorCURIE):
            self.source = URIorCURIE(self.source)

        if not isinstance(self.pipeline, list):
            self.pipeline = [self.pipeline] if self.pipeline is not None else []
        self.pipeline = [v if isinstance(v, LexicalTransformationPipelineName) else LexicalTransformationPipelineName(v) for v in self.pipeline]

        super().__post_init__(**kwargs)


class Activity(YAMLRoot):
    """
    Generic grouping for any lexical operation
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = PROV.Activity
    class_class_curie: ClassVar[str] = "prov:Activity"
    class_name: ClassVar[str] = "Activity"
    class_model_uri: ClassVar[URIRef] = MRULES.Activity


@dataclass
class LexicalTransformationPipeline(Activity):
    """
    A collection of atomic lexical transformations that are applied in serial fashion
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LI.LexicalTransformationPipeline
    class_class_curie: ClassVar[str] = "li:LexicalTransformationPipeline"
    class_name: ClassVar[str] = "LexicalTransformationPipeline"
    class_model_uri: ClassVar[URIRef] = MRULES.LexicalTransformationPipeline

    name: Union[str, LexicalTransformationPipelineName] = None
    transformations: Optional[Union[Union[dict, "LexicalTransformation"], List[Union[dict, "LexicalTransformation"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, LexicalTransformationPipelineName):
            self.name = LexicalTransformationPipelineName(self.name)

        if not isinstance(self.transformations, list):
            self.transformations = [self.transformations] if self.transformations is not None else []
        self.transformations = [v if isinstance(v, LexicalTransformation) else LexicalTransformation(**as_dict(v)) for v in self.transformations]

        super().__post_init__(**kwargs)


@dataclass
class LexicalTransformation(Activity):
    """
    An atomic lexical transformation applied on a term (string) yielding a transformed string
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LI.LexicalTransformation
    class_class_curie: ClassVar[str] = "li:LexicalTransformation"
    class_name: ClassVar[str] = "LexicalTransformation"
    class_model_uri: ClassVar[URIRef] = MRULES.LexicalTransformation

    type: Optional[Union[str, "TransformationType"]] = None
    params: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.type is not None and not isinstance(self.type, TransformationType):
            self.type = TransformationType(self.type)

        if self.params is not None and not isinstance(self.params, str):
            self.params = str(self.params)

        super().__post_init__(**kwargs)


# Enumerations
class TransformationType(EnumDefinitionImpl):
    """
    A controlled datamodels of the types of transformation that can be applied to
    """
    Stemming = PermissibleValue(text="Stemming",
                                       description="Removal of the last few characters of a word to yield a stem term for each word in the term")
    Lemmatization = PermissibleValue(text="Lemmatization",
                                                 description="Contextual reduction of a word to its base form for each word in the term")
    WordOrderNormalization = PermissibleValue(text="WordOrderNormalization",
                                                                   description="reorder words in the term to a standard order such that comparisons are order-independent")
    Depluralization = PermissibleValue(text="Depluralization",
                                                     description="Transform plural form to singular form for each word in a term")
    CaseNormalization = PermissibleValue(text="CaseNormalization",
                                                         description="Transform term to a standard case, typically lowercase")
    WhitespaceNormalization = PermissibleValue(text="WhitespaceNormalization",
                                                                     description="Trim whitespace, condense whitespace runs, and transform all non-space whitespace to spaces")
    TermExpanson = PermissibleValue(text="TermExpanson",
                                               description="Expand terms using a dictionary")

    _defn = EnumDefinition(
        name="TransformationType",
        description="A controlled datamodels of the types of transformation that can be applied to",
    )

# Slots
class slots:
    pass

slots.mappingRuleCollection__rules = Slot(uri=MRULES.rules, name="mappingRuleCollection__rules", curie=MRULES.curie('rules'),
                   model_uri=MRULES.mappingRuleCollection__rules, domain=None, range=Optional[Union[Union[dict, MappingRule], List[Union[dict, MappingRule]]]])

slots.mappingRuleCollection__minimum_confidence = Slot(uri=MRULES.minimum_confidence, name="mappingRuleCollection__minimum_confidence", curie=MRULES.curie('minimum_confidence'),
                   model_uri=MRULES.mappingRuleCollection__minimum_confidence, domain=None, range=Optional[float])

slots.mappingRule__description = Slot(uri=MRULES.description, name="mappingRule__description", curie=MRULES.curie('description'),
                   model_uri=MRULES.mappingRule__description, domain=None, range=Optional[str])

slots.mappingRule__oneway = Slot(uri=MRULES.oneway, name="mappingRule__oneway", curie=MRULES.curie('oneway'),
                   model_uri=MRULES.mappingRule__oneway, domain=None, range=Optional[Union[bool, Bool]])

slots.mappingRule__preconditions = Slot(uri=SH.condition, name="mappingRule__preconditions", curie=SH.curie('condition'),
                   model_uri=MRULES.mappingRule__preconditions, domain=None, range=Optional[Union[dict, Precondition]])

slots.mappingRule__postconditions = Slot(uri=MRULES.postconditions, name="mappingRule__postconditions", curie=MRULES.curie('postconditions'),
                   model_uri=MRULES.mappingRule__postconditions, domain=None, range=Optional[Union[dict, Postcondition]])

slots.precondition__subject_source_one_of = Slot(uri=MRULES.subject_source_one_of, name="precondition__subject_source_one_of", curie=MRULES.curie('subject_source_one_of'),
                   model_uri=MRULES.precondition__subject_source_one_of, domain=None, range=Optional[Union[str, List[str]]])

slots.precondition__object_source_one_of = Slot(uri=MRULES.object_source_one_of, name="precondition__object_source_one_of", curie=MRULES.curie('object_source_one_of'),
                   model_uri=MRULES.precondition__object_source_one_of, domain=None, range=Optional[Union[str, List[str]]])

slots.precondition__mapping_source_one_of = Slot(uri=MRULES.mapping_source_one_of, name="precondition__mapping_source_one_of", curie=MRULES.curie('mapping_source_one_of'),
                   model_uri=MRULES.precondition__mapping_source_one_of, domain=None, range=Optional[Union[str, List[str]]])

slots.precondition__subject_match_field_one_of = Slot(uri=MRULES.subject_match_field_one_of, name="precondition__subject_match_field_one_of", curie=MRULES.curie('subject_match_field_one_of'),
                   model_uri=MRULES.precondition__subject_match_field_one_of, domain=None, range=Optional[Union[str, List[str]]])

slots.precondition__object_match_field_one_of = Slot(uri=MRULES.object_match_field_one_of, name="precondition__object_match_field_one_of", curie=MRULES.curie('object_match_field_one_of'),
                   model_uri=MRULES.precondition__object_match_field_one_of, domain=None, range=Optional[Union[str, List[str]]])

slots.precondition__transformations_included_in = Slot(uri=MRULES.transformations_included_in, name="precondition__transformations_included_in", curie=MRULES.curie('transformations_included_in'),
                   model_uri=MRULES.precondition__transformations_included_in, domain=None, range=Optional[Union[str, List[str]]])

slots.postcondition__predicate_id = Slot(uri=MRULES.predicate_id, name="postcondition__predicate_id", curie=MRULES.curie('predicate_id'),
                   model_uri=MRULES.postcondition__predicate_id, domain=None, range=Optional[str])

slots.postcondition__weight = Slot(uri=MRULES.weight, name="postcondition__weight", curie=MRULES.curie('weight'),
                   model_uri=MRULES.postcondition__weight, domain=None, range=Optional[float])

slots.lexicalIndex__groupings = Slot(uri=MRULES.groupings, name="lexicalIndex__groupings", curie=MRULES.curie('groupings'),
                   model_uri=MRULES.lexicalIndex__groupings, domain=None, range=Optional[Union[Dict[Union[str, LexicalGroupingTerm], Union[dict, LexicalGrouping]], List[Union[dict, LexicalGrouping]]]])

slots.lexicalIndex__pipelines = Slot(uri=MRULES.pipelines, name="lexicalIndex__pipelines", curie=MRULES.curie('pipelines'),
                   model_uri=MRULES.lexicalIndex__pipelines, domain=None, range=Optional[Union[Dict[Union[str, LexicalTransformationPipelineName], Union[dict, LexicalTransformationPipeline]], List[Union[dict, LexicalTransformationPipeline]]]])

slots.lexicalGrouping__term = Slot(uri=MRULES.term, name="lexicalGrouping__term", curie=MRULES.curie('term'),
                   model_uri=MRULES.lexicalGrouping__term, domain=None, range=URIRef)

slots.lexicalGrouping__relationships = Slot(uri=MRULES.relationships, name="lexicalGrouping__relationships", curie=MRULES.curie('relationships'),
                   model_uri=MRULES.lexicalGrouping__relationships, domain=None, range=Optional[Union[Union[dict, RelationshipToTerm], List[Union[dict, RelationshipToTerm]]]])

slots.relationshipToTerm__predicate = Slot(uri=MRULES.predicate, name="relationshipToTerm__predicate", curie=MRULES.curie('predicate'),
                   model_uri=MRULES.relationshipToTerm__predicate, domain=None, range=Optional[Union[str, URIorCURIE]])

slots.relationshipToTerm__element = Slot(uri=MRULES.element, name="relationshipToTerm__element", curie=MRULES.curie('element'),
                   model_uri=MRULES.relationshipToTerm__element, domain=None, range=Optional[Union[str, URIorCURIE]])

slots.relationshipToTerm__element_term = Slot(uri=MRULES.element_term, name="relationshipToTerm__element_term", curie=MRULES.curie('element_term'),
                   model_uri=MRULES.relationshipToTerm__element_term, domain=None, range=Optional[str])

slots.relationshipToTerm__source = Slot(uri=MRULES.source, name="relationshipToTerm__source", curie=MRULES.curie('source'),
                   model_uri=MRULES.relationshipToTerm__source, domain=None, range=Optional[Union[str, URIorCURIE]])

slots.relationshipToTerm__pipeline = Slot(uri=MRULES.pipeline, name="relationshipToTerm__pipeline", curie=MRULES.curie('pipeline'),
                   model_uri=MRULES.relationshipToTerm__pipeline, domain=None, range=Optional[Union[Union[str, LexicalTransformationPipelineName], List[Union[str, LexicalTransformationPipelineName]]]])

slots.lexicalTransformationPipeline__name = Slot(uri=MRULES.name, name="lexicalTransformationPipeline__name", curie=MRULES.curie('name'),
                   model_uri=MRULES.lexicalTransformationPipeline__name, domain=None, range=URIRef)

slots.lexicalTransformationPipeline__transformations = Slot(uri=MRULES.transformations, name="lexicalTransformationPipeline__transformations", curie=MRULES.curie('transformations'),
                   model_uri=MRULES.lexicalTransformationPipeline__transformations, domain=None, range=Optional[Union[Union[dict, LexicalTransformation], List[Union[dict, LexicalTransformation]]]])

slots.lexicalTransformation__type = Slot(uri=MRULES.type, name="lexicalTransformation__type", curie=MRULES.curie('type'),
                   model_uri=MRULES.lexicalTransformation__type, domain=None, range=Optional[Union[str, "TransformationType"]])

slots.lexicalTransformation__params = Slot(uri=MRULES.params, name="lexicalTransformation__params", curie=MRULES.curie('params'),
                   model_uri=MRULES.lexicalTransformation__params, domain=None, range=Optional[str])



from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.interfaces.ontology_interface import OntologyInterface
from oaklib.interfaces.validator_interface import ValidatorInterface
from oaklib.interfaces.subsetter_interface import SubsetterInterface
from oaklib.interfaces.relation_graph_interface import RelationGraphInterface



"""
Datamodels
----------

..note ::

    this package includes multiple alternative overlapping datamodels;
    in general a maximum of one datamodel is required for any one use case.

"""
__version__ = '0.1.0'

from oaklib.interfaces import BasicOntologyInterface




from dataclasses import dataclass
from typing import List, Optional

from oaklib.types import URI

VAR_NAME = str
WHERE_CLAUSE = str


@dataclass
class SparqlQuery:
    """
    Represents a SPARQL query
    """
    distinct: bool = None
    select: List[VAR_NAME] = None
    graph: Optional[URI] = None
    where: List[WHERE_CLAUSE] = None

    def select_str(self):
        distinct = 'DISTINCT ' if self.distinct else ''
        return f'{distinct}{" ".join(self.select)} '

    def where_str(self):
        return ". ".join([w for w in self.where if w])

    def query_str(self):
        """
        Generate the SPARQL query string
        :return:
        """
        w = self.where_str()
        if self.graph:
            w = f'GRAPH <{self.graph}> {{ {w} }}'
        return f'SELECT {self.select_str()} WHERE {{ {w} }}'


import sys
from dataclasses import dataclass, field
from typing import Any


@dataclass
class StreamingWriter:
    """
    Base class for streaming writers
    """
    file: Any = field(default_factory=lambda: sys.stdout)

    def close(self):
        pass


import logging
import os
from pathlib import Path
from appdirs import user_data_dir
from oaklib.datamodels.vocabulary import APP_NAME

APIKEY_SUFFIX = 'apikey.txt'

def get_apikey_path(system: str) -> Path:
    p = Path(user_data_dir(APP_NAME)) / f'{system}-{APIKEY_SUFFIX}'
    logging.info(f'API KEY path = {p}')
    return p

def get_apikey_value(system: str) -> str:
    path = get_apikey_path(system)
    if not path.exists():
        raise ValueError(f'No API key found in: {path}')
    with open(path) as stream:
        return stream.readlines()[0].strip()

def set_apikey_value(system: str, val: str) -> str:
    dir = Path(user_data_dir(APP_NAME))
    dir.mkdir(exist_ok=True, parents=True)
    with open(get_apikey_path(system), 'w', encoding='utf-8') as stream:
        stream.write(val)




"""
Utilities for traversing Ontology Graphs
====================


"""
from copy import copy
from dataclasses import dataclass
from typing import List, Union, Dict, Iterable, Tuple

from oaklib.interfaces.basic_ontology_interface import RELATIONSHIP, BasicOntologyInterface
from oaklib.types import CURIE, PRED_CURIE

#PATH = Tuple[CURIE, List[PRED_CURIE], CURIE]
PATH = List[RELATIONSHIP]


def walk_up(oi: BasicOntologyInterface, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[RELATIONSHIP]:
    """
    Walks up the relation graph from a seed set of curies or individual curie, returning the full ancestry graph

    Note: this may be inefficient for remote endpoints, in future a graph walking endpoint will implement this

    :param oi:
    :param start_curies:
    :param predicates:
    :return:
    """
    if isinstance(start_curies, CURIE):
        next_curies = [start_curies]
    else:
        next_curies = copy(start_curies) # do not mutate
    rels = []
    visited = copy(next_curies)
    while len(next_curies) > 0:
        next_curie = next_curies.pop()
        for pred, fillers in oi.get_outgoing_relationships_by_curie(next_curie).items():
            if not predicates or pred in predicates:
                for filler in fillers:
                    if filler not in visited:
                        next_curies.append(filler)
                        visited.append(filler)
                    rels.append((next_curie, pred, filler))
    for rel in rels:
        yield rel


def walk_down(oi: BasicOntologyInterface, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[RELATIONSHIP]:
    """
    As walk_up, but traversing incoming, not outgoing relationships

    :param oi:
    :param start_curies:
    :param predicates:
    :return:
    """
    if isinstance(start_curies, CURIE):
        next_curies = [start_curies]
    else:
        next_curies = copy(start_curies) # do not mutate
    rels = []
    visited = copy(next_curies)
    while len(next_curies) > 0:
        next_curie = next_curies.pop()
        for pred, subjects in oi.get_incoming_relationships_by_curie(next_curie).items():
            if not predicates or pred in predicates:
                for subject in subjects:
                    if subject not in visited:
                        next_curies.append(subject)
                        visited.append(subject)
                    rels.append((subject, pred, next_curie))
    for rel in rels:
        yield rel



from abc import ABC
from typing import Dict, List, Tuple, Iterable

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
from oaklib.types import CURIE, LABEL, URI
from rdflib import Graph


class SkosInterface(BasicOntologyInterface, ABC):
    """
    presents ontology as simple SKOS vocabularies

    `SKOS <https://www.w3.org/2004/02/skos/>`_
    """

    def concepts(self) -> Iterable[CURIE]:
        """

        :return:
        """
        raise NotImplementedError

    def broader(self, curie: CURIE) -> Iterable[CURIE]:
        """

        :param curie:
        :return:
        """
        raise NotImplementedError

    def narrower(self, curie: CURIE) -> Iterable[CURIE]:
        """

        :param curie:
        :return:
        """
        raise NotImplementedError

    def exact(self, curie: CURIE) -> Iterable[CURIE]:
        """

        :param curie:
        :return:
        """
        raise NotImplementedError

    def related(self, curie: CURIE) -> Iterable[CURIE]:
        """

        :param curie:
        :return:
        """
        raise NotImplementedError


